import hcl2
import lark.exceptions
from lark import UnexpectedCharacters


class LinesMetric:
    """ This is an abstract class the concrete classes measuring lines of code will extend.
    """

    def __init__(self, script: str):
        """The class constructor.

        Parameters
        ----------
        script : str
            A plain Terraform file

        """

        if script is None:
            raise TypeError("Parameter 'script' meant to be a string, not None.")
        try :
            # Check if is a valid hcl file
            self.__hcl = hcl2.loads(script, with_meta=True)
            if self.__hcl is None:
                raise TypeError("Expected a not empty Terraform script")

            self.__hcl = script

        except UnexpectedCharacters as e:
            raise TypeError("Expected a valid Terraform script")
        except lark.exceptions.VisitError:
            self.__hcl = script

    @property
    def hcl(self):
        """Il file Terraform in formato HCL."""
        return self.__hcl

    def count(self):
        """Metodo per eseguire la metrica."""
        pass


from terraformmetrics.terraform_metric import TerraformMetric

import re


class NumRegex(TerraformMetric):
    """ This class measures the number of regex used, counting the number of regex and regexall functions
    """

    def count(self):
        """Return the number of regex or regexall functions
        Returns
        -------
        int
            Number of regex or regexall
        """

        # Funzione ricorsiva per contare il numero di funzioni regex o regexall
        def count_regex(data):
            count = 0
            if isinstance(data, dict):
                for key, value in data.items():
                    count += count_regex(value)
            elif isinstance(data, list):
                for item in data:
                    count += count_regex(item)
            elif isinstance(data, str):
                if re.search(r'regex\(', data) or re.search(r'regexall\(', data):
                    count += 1
            return count

        total_regex_count = count_regex(self.tfparsed)

        return total_regex_count



import argparse
import os
import json

import lark

from terraformmetrics.metrics_extractor import extract_all

VERSION = "1.0.3"


def get_parser():
    description = 'Extract metrics from Terraform scripts.'

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(action='store', dest='src', help='source file (playbook or tasks file) or directory')
    parser.add_argument('--omit-zero-metrics', dest='omit_zero_metrics', action='store_true',
                        help='omit metrics with value equal 0')
    parser.add_argument('-d', '--dest', help='destination path to save results')
    parser.add_argument('-o', '--output', action='store_true', help='shows output')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + VERSION)

    return parser


def load_file(path: str) -> str:
    """ Returns a String object representing the content of the file at <path>, if any; None otherwise """
    if os.path.isfile(path):
        content = ''
        with open(path, 'r') as file:
            content = file.read()

        return content


def load_dir(input_dir: str) -> dict:
    """
    Find all terraform file in the input directory and return a dictionary of k:v pairs where k is the filepath and \
        the v the content of the file at that path
    """

    dirs_stack = []

    for item in os.listdir(input_dir):
        dirs_stack.append(os.path.join(input_dir, item))

    tf_files = []

    while len(dirs_stack) > 0:
        item = dirs_stack.pop()

        if os.path.isfile(item):
            _, file_extension = os.path.splitext(item)

            if file_extension in ['.tf']:
                tf_files.append(item)

        elif os.path.isdir(item):
            for sub_item in os.listdir(item):
                dirs_stack.append(os.path.join(item, sub_item))

    scripts = dict()

    for file_path in tf_files:
        scripts[file_path] = load_file(file_path)

    return scripts


# funzione per caricare la risorsa di partenza
def load_source(path: str) -> dict:
    if os.path.isfile(path):
        return {path: load_file(path)}
    elif os.path.isdir(path):
        return load_dir(path)

    return dict()


#funzione che permette di salvare il risultato
def save_output(json_data, out_path):
    if not os.path.isfile(out_path):
        data = []
    else:
        with open(out_path, 'r') as json_file:
            data = json.load(json_file)

    data.append(json_data)

    with open(out_path, 'w') as out_file:
        json.dump(data, out_file, sort_keys=True)


def cli():
    args = get_parser().parse_args()

    scripts = load_source(args.src)
    for file_path, script in scripts.items():

        try:
            metrics = extract_all(script)
            metrics['filepath'] = file_path

            if args.omit_zero_metrics:  # Show only non-zero metrics
                metrics = {k: v for k, v in metrics.items() if v != 0}

            if args.dest:
                save_output(metrics, args.dest)

            # Pretty-print json to screen (always print if args.dest in not defined)
            if args.output or not args.dest:
                metrics_2_json = json.dumps(metrics, indent=4, sort_keys=True)
                print(metrics_2_json)

        #except TypeError:
         # print('\033[91m' + f'{file_path} is not a valid HCL2 file.' + '\033[0m')
        except Exception as e:
            print(f'An unknown error has occurred: {str(e)}')



from terraformmetrics.terraform_metric import TerraformMetric

import re


class NumMath(TerraformMetric):
    """ This class measures the number of math operators [' + ', ' - ', ' * ', ' / ', ' % ', ' ^ '] in a Script
    """

    def count(self):
        """Return the number of math operators
        Returns
        -------
        int
            Number of math operators
        """

        # Funzione ricorsiva per contare il numero di funzioni try
        def count_operators(expression):
            if isinstance(expression, list):
                count = 0
                for item in expression:
                    count += count_operators(item)
                return count
            elif isinstance(expression, dict):
                count = 0
                for key, value in expression.items():
                    if isinstance(value, str):
                        operators = [' + ', ' - ', ' * ', ' / ', ' % ', ' ^ ']
                        count += sum(1 for operator in operators if operator in value)
                    count += count_operators(value)
                return count
            else:
                return 0

        total_try_count = count_operators(self.tfparsed)

        return total_try_count


from terraformmetrics.terraform_metric import TerraformMetric


class NumDataSources(TerraformMetric):
    """ This class measures the number of data sources
    """

    def count(self):
        """Return the number of data sources used
        Returns
        -------
        int
            Number of data sources used
        """

        num_data = 0

        if 'data' in self.tfparsed:
            num_data = len(self.tfparsed['data'])

        return num_data


from terraformmetrics.lines_metric import LinesMetric

class LinesComment(LinesMetric):
    """ This class measures the number of comments in a Terraform script"""

    def count(self):
        """Return the number of commented lines
        Returns
        -------
        int
            Number of commented lines
        """
        cloc = 0

        in_multiline_comment = False
        for line in self.hcl.splitlines():
            stripped_line = line.strip()
            if not stripped_line:
                continue
            if stripped_line.startswith("#"):
                cloc += 1

            # Gestisci commenti su più linee
            if stripped_line.startswith("/*"):
                in_multiline_comment = True
                cloc += 1
                if "*/" in stripped_line:
                    in_multiline_comment = False
                    continue
                else:
                    continue
            elif "*/" in stripped_line:
                in_multiline_comment = False
                cloc += 1
                continue

            if in_multiline_comment:
                cloc += 1

        return cloc



from terraformmetrics.terraform_metric import TerraformMetric


class NumDynamicBlock(TerraformMetric):
    """ This class measures the number of dynamic blocks in a Script Terraform
    """

    def count(self):
        """Return the number of dynamic blocks
        Returns
        -------
        int
            Number of dynamics
        """

        # Funzione ricorsiva per contare il numero di blocchi dynamic all'interno di un file
        def count_dynamic_blocks(data):
            count = 0

            if isinstance(data, dict):
                for key, value in data.items():
                    if key == "dynamic":
                        count += len(value)
                    else:
                        count += count_dynamic_blocks(value)
            elif isinstance(data, list):
                for item in data:
                    count += count_dynamic_blocks(item)

            return count

        total_dynamic_count = count_dynamic_blocks(self.tfparsed)

        return total_dynamic_count



from terraformmetrics.terraform_metric import TerraformMetric


class NumVar(TerraformMetric):
    """ This class measures the number of input vars in a Script Terraform
    """

    def count(self):
        """Return the number of input vars
        Returns
        -------
        int
            Number of input vars
        """

        # Funzione ricorsiva per contare il numero di variabili all'interno di un file
        def count_variables(data):
            count = 0

            if isinstance(data, dict):
                for key, value in data.items():
                    if key == 'variable' and isinstance(value, list):
                        for variable in value:
                            count += 1
            elif isinstance(data, list):
                for item in data:
                    count += count_variables(item)

            return count

        total_var_count = count_variables(self.tfparsed)

        return total_var_count




from terraformmetrics.terraform_metric import TerraformMetric


class AvgResourcesSize(TerraformMetric):
    """ This class measures the average number of lines of code of resources
    """

    def count(self):
        """Return the average resource size
        Returns
        -------
        int
            Average resource size, rounded to the nearest unit
        """

        def find_start_end_lines(d):
            if isinstance(d, dict):
                if '__start_line__' in d and '__end_line__' in d:
                    return d['__start_line__'], d['__end_line__']
                for value in d.values():
                    lines = find_start_end_lines(value)
                    if lines:
                        return lines
            return None

        total_resources_lines = 0
        avg_resource_size = 0

        resource_list = self.tfparsed.get('resource', [])
        for resource in resource_list:
            start_line, end_line = find_start_end_lines(resource)

            empty_line_count = 0

            for line_number, line in enumerate(self.tfscript, start=1):
                if start_line <= line_number <= end_line:
                    if not line.strip():
                        empty_line_count += 1

            total_lines = end_line - start_line + 1
            source_lines = total_lines - empty_line_count

            total_resources_lines += source_lines
        if len(resource_list) != 0:
            avg_resource_size = int(round(total_resources_lines/len(resource_list)))
        return avg_resource_size



import re
from terraformmetrics.terraform_metric import TerraformMetric

COMPARISON_OPERATORS = re.compile(r'(?<= )&&(?= )|(?<= )!!(?= )|(?<= )!(?= )')


class NumDecisions(TerraformMetric):
    """ This class measures the number of decisions within an if block in a Script Terraform

    decisions are identified by counting the number of occurrences of the operators : (&& || !)
    within an if block
    """

    def count(self):
        """Return the number of decisions
        Returns
        -------
        int
            Number of decisions
        """

        # Funzione ricorsiva per contare operatori all'interno di un dizionario o una lista
        def count_operators(data):
            count = 0
            if isinstance(data, dict):
                for key, value in data.items():
                    count += count_operators(value)
            elif isinstance(data, list):
                for item in data:
                    count += count_operators(item)
            elif isinstance(data, str):
                if re.search(r' \? ', data) and re.search(r' : ', data):
                    no_quotes_text = re.sub(r'"(.*?)"', ' ', data)
                    count += len(COMPARISON_OPERATORS.findall(str(no_quotes_text)))
            return count

        total_operator_count = count_operators(self.tfparsed)

        return total_operator_count


#General Metrics
from terraformmetrics.general.lines_code import LinesCode
from terraformmetrics.general.lines_blank import LinesBlank
from terraformmetrics.general.lines_comment import LinesComment
from terraformmetrics.general.num_conditions import NumConditions
from terraformmetrics.general.num_decisions import NumDecisions
from terraformmetrics.general.num_tokens import NumTokens
from terraformmetrics.general.num_math_operations import NumMath
from terraformmetrics.general.text_entropy import TextEntropy

#Terraform specific
from terraformmetrics.terraformspec.num_resources import NumResources
from terraformmetrics.terraformspec.avg_resource_size import AvgResourcesSize
from terraformmetrics.terraformspec.num_remote_exec import NumRemoteExec
from terraformmetrics.terraformspec.num_file_provisioner import NumFileProvisioner
from terraformmetrics.terraformspec.num_permissions import NumFilePermissions
from terraformmetrics.terraformspec.num_modules import NumModules
from terraformmetrics.terraformspec.num_http_datasource import NumHttp
from terraformmetrics.terraformspec.num_error_handling import NumErrorHandling
from terraformmetrics.terraformspec.num_regex import NumRegex
from terraformmetrics.terraformspec.num_path import NumPath
from terraformmetrics.terraformspec.num_var import NumVar
from terraformmetrics.terraformspec.num_output import NumOutputs
from terraformmetrics.terraformspec.num_provider import NumProviders
from terraformmetrics.terraformspec.num_depends_on import NumDependsOn
from terraformmetrics.terraformspec.num_data_sources import NumDataSources
from terraformmetrics.terraformspec.num_dynamic_blocks import NumDynamicBlock
from terraformmetrics.terraformspec.num_local import NumLocals

general_metrics = {
                    'lines_code': LinesCode,
                    'lines_blank': LinesBlank,
                    'lines_comment': LinesComment,
                    'num_conditions': NumConditions,
                    'num_decisions': NumDecisions,
                    'num_math': NumMath,
                  }
tf_metrics = {
                'num_resources': NumResources,
                'avg_resources_size': AvgResourcesSize,
                'num_remote_exec_provisioner': NumRemoteExec,
                'num_file_provisioner': NumFileProvisioner,
                'num_permissions': NumFilePermissions,
                'num_modules': NumModules,
                'num_http_datasource': NumHttp,
                'num_error_handling': NumErrorHandling,
                'num_tokens': NumTokens,
                'num_regex': NumRegex,
                'num_path': NumPath,
                'num_var': NumVar,
                'num_outputs': NumOutputs,
                'num_providers': NumProviders,
                'num_depends_on': NumDependsOn,
                'num_data_sources': NumDataSources,
                'num_dynamic': NumDynamicBlock,
                'num_locals': NumLocals,
                'text_entropy' : TextEntropy,
             }




from terraformmetrics.lines_metric import LinesMetric


class LinesCode(LinesMetric):
    """ This class measures the number of executable lines of code in a Terraform script.
    """

    def count(self):
        """Return the number of executable lines of code.

        Returns
        -------
        int
            Number of lines of code
        """
        loc = 0
        in_multiline_comment = False
        for line in self.hcl.splitlines():
            stripped_line = line.strip()
            if not stripped_line:
                continue
            if stripped_line.startswith("#"):
                continue

            # Gestisci commenti su più linee
            if stripped_line.startswith("/*"):
                in_multiline_comment = True
                if "*/" in stripped_line:
                    in_multiline_comment = False
                    continue
                else:
                    continue
            elif "*/" in stripped_line:
                in_multiline_comment = False
                continue

            if in_multiline_comment:
                continue

            loc += 1

        return loc



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = '1.0.3'

setup(name='terraformmetrics',
      version=VERSION,
      description='A module to extract metrics from Terrafor, scripts',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Gerardo Brescia',
      maintainer='Gerardo Brescia',
      author_email='gerrybrescia99@gmail.com',
      url='https://github.com/GerardoBrescia/radon-terraform-metrics.git',
      download_url=f'https://github.com/radon-h2020/radon-ansible-metrics/archive/{VERSION}.tar.gz',
      packages=find_packages(exclude=('test',)),
      entry_points = {
        'console_scripts': ['terraform-metrics=terraformmetrics.command_line:cli'],
      },
      python_requires='>=3.6',
      classifiers=[
         "Development Status :: 5 - Production/Stable",
         "Intended Audience :: Developers",
         "Programming Language :: Python :: 3.6",
         "Programming Language :: Python :: 3.7",
         "Programming Language :: Python :: 3.8",
         "Programming Language :: Python :: 3.9",
         "License :: OSI Approved :: Apache Software License",
         "Topic :: Software Development :: Libraries :: Python Modules",
         "Operating System :: OS Independent"
      ],
      install_requires=[
        'python-hcl2==4.3.2'
      ],
      test_requirements=[
        'pytest==5.4.2'
      ]
)



from terraformmetrics.terraform_metric import TerraformMetric


class NumOutputs(TerraformMetric):
    """ This class measures the number of outputs used
    """

    def count(self):
        """Return the number of outputs used
        Returns
        -------
        int
            Number of outputs used
        """

        num_outputs = 0

        if 'output' in self.tfparsed:
            num_outputs = len(self.tfparsed['output'])

        return num_outputs



from math import log2

from terraformmetrics.terraform_metric import TerraformMetric

import re


class TextEntropy(TerraformMetric):
    """ This class measures the text entropy
    """

    def count(self):
        """Return the text entropy
        Returns
        -------
        float
            text entropy
        """

        def get_words(data):
            words= []
            if isinstance(data, dict):
                for key, value in data.items():
                    if key != '__start_line__' and key != '__end_line__' and key != '__len__':
                        if isinstance(value, list):
                            for item in value:
                                words.append(key)
                        else:
                            words.append(key)
                        remaining_words = get_words(value)
                        words.extend(remaining_words)
            elif isinstance(data, list):
                for item in data:
                    if data != '__start_line__' and data != '__end_line__' and data != '__len__':
                        words.extend(get_words(item))
            elif isinstance(data, str):
                if data != '__start_line__' and data != '__end_line__' and data != '__len__':
                    words.extend(str(data).split())
            return words

        words = get_words(self.tfparsed)

        words_set = set(words)

        freq = {word: words.count(word) for word in words_set}

        entropy = 0
        for word in words_set:
            p = freq[word] / len(words)
            entropy -= p * log2(p)

        return round(entropy, 2)
